﻿// Copyright © 2014 Corel Corporation.
// 
// Permission is hereby granted, free of charge, to any person or organization obtaining a copy of the software and accompanying 
// documentation covered by this license (the "Software") to use, reproduce, display, distribute, execute, and transmit the Software, 
// and to prepare derivative works of the Software, and to permit third-parties to whom the Software is furnished to do so, all 
// subject to the following:
// 
// The copyright notices in the Software and this entire statement, including the above license grant, the original location it was 
// downloaded from, this restriction and the following disclaimer, must be included in all copies of the Software, in whole or in part, 
// and all derivative works of the Software, unless such copies or derivative works are solely in the form of machine-executable object 
// code generated by a source language processor.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND NON-INFRINGEMENT. THE SOFTWARE MAY CONTAIN BUGS, ERRORS AND OTHER 
// PROBLEMS THAT COULD CAUSE SYSTEM FAILURES AND THE USE OF SUCH SOFTWARE IS ENTIRELY AT THE USER’S RISK. IN NO EVENT SHALL THE 
// COPYRIGHT HOLDERS OR ANYONE DISTRIBUTING THE SOFTWARE BE LIABLE FOR ANY DAMAGES OR OTHER LIABILITY, WHETHER IN CONTRACT, TORT OR 
// OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Corel.Interop.VGCore;

namespace CGS
{
	///<summary>
	/// ZigZagTool: Simple 2-point line tool example.  Implements ToolState which allows CorelDRAW to call back into the tool
	///</summary>
	public class ZigZagTool : ToolState
	{
		///<summary>
		/// Constructor for the tool.  Keep a pointer to the application object
		///</summary>
		public ZigZagTool(Application App)
		{
			Application = App;  // Save the application, we use it to create the new line segment
			Application.RegisterUserApplicationPreference("ZigZagTool", "Height", 0.125);
			Application.RegisterUserApplicationPreference("ZigZagTool", "Distance", 0.25);
		}

		///<summary>
		/// OnStartState: Called when the tool is started
		///</summary>
		public void OnStartState(ToolStateAttributes StateAttributes)
		{
			// save the state attributes so that we can reuse it later
			currentAttribs = StateAttributes;
			currentAttribs.SetCursor(cdrCursorShape.cdrCursorSmallcrosshair);
			currentAttribs.AllowTempPickState = false;
			currentAttribs.PropertyBarGuid = "4c09c866-b21f-4e04-9c35-e0dae7b4ad90";
			ui = Application.CreateOnScreenCurve(); // create the xor'd UI object
			pts = Application.Math.CreatePointRange();
		}

		///<summary>
		/// OnExitState: Called when the tool is exited
		///</summary>
		public void OnExitState()
		{
			Reset(); // cleanup and reset UI and variable
		}

		///<summary>
		/// IsDrawing: Called to determine if the tool is in a drawing state.  Certain operation, like panning, 
		/// will function differently when this returns true.
		///</summary>
		public bool IsDrawing
		{
			get { return isDrawing; }
		}

		///<summary>
		/// OnAbort: Called when the user presses escape
		///</summary>
		public void OnAbort()
		{
			Reset(); // cleanup and reset UI and variable
		}

		///<summary>
		/// OnCommit: Called when the user presses enter, or in this case, by OnLButtonUp when the left mouse button 
		/// is released
		///</summary>
		public void OnCommit(Point pt)
		{
			if (isDrawing)
			{
				Curve curve = CreateZigZagCurve();
				Application.ActiveLayer.CreateCurve(curve);
				Reset(); // cleanup and reset UI and variable
			}
		}

		///<summary>
		/// OnLButtonDownLeaveGrace: This is called when the left mouse button is held in place for a short period of 
		/// time, or when the left mouse button is held and the mouse moves a short distance. 'pt' will contain the
		/// point when the mouse button was pressed.
		///</summary>
		public void OnLButtonDownLeaveGrace(Point pt)
		{
			isDrawing = true; // place us into the drawing substate
			pts.AddPoint(pt);
		}

		///<summary>
		/// OnLButtonUp: Called when the user releases the left mouse button
		///</summary>
		public void OnLButtonUp(Point pt)
		{
			OnCommit(pt); // forward to OnCommit, which will create the line segment
		}

		///<summary>
		/// OnMouseMove: Called when the mouse moves across the view.
		///</summary>
		public void OnMouseMove(Point pt)
		{
			if (isDrawing)
			{
				pts.AddPoint(pt);
				// update the xor'd line
				Curve curve = CreateZigZagCurve();
				if (curve != null)
				{
					ui.SetCurve(curve);
					ui.Show();
				}
			}
		}

		///<summary>
		/// OnSnapMouse: Called for every mouse move, this allows us to adjust the mouse point before it reaches any 
		/// of the other functions.  The currentAttribs has helper functions to use the built in snapping.
		///</summary>
		public void OnSnapMouse(Point pt, ref bool Handled)
		{
			Handled = isDrawing; // use the default snapping if we are not drawing
		}

		///<summary>
		/// OnDelete: Called when the delete key is pressed.  We do nothing to use the default bahaviour
		///</summary>
		public void OnDelete(ref bool Handled) { }

		///<summary>
		/// OnKeyDown: Called when a key is pressed. We do nothing to use the default bahaviour
		///</summary>
		public void OnKeyDown(int windowsKeyCode, ref bool Handled) { }

		///<summary>
		/// OnKeyUp: Called when a key is released.  We do nothing to use the default bahaviour
		///</summary>
		public void OnKeyUp(int windowsKeyCode, ref bool Handled) { }

		///<summary>
		/// OnLButtonDblClick: Called when the left mouse button is double-clicked.  We do nothing to use 
		/// the default bahaviour
		///</summary>
		public void OnLButtonDblClick(Point pt) { }

		///<summary>
		/// OnLButtonDown: Called when the left mouse button is pressed.  We do nothing because the left mouse button
		/// behaviour is handled in OnLButtonDownLeaveGrace.
		///</summary>
		public void OnLButtonDown(Point pt) { }

		///<summary>
		/// OnClick: Called when the left mouse button is clicked.  We do nothing to use the default bahaviour
		/// The default behaviour is to select the shape beneath the mouse cursor.
		///</summary>
		public void OnClick(Point pt, ref bool Handled) { }

		///<summary>
		/// OnRButtonDown: Called when the right mouse button is pressed.  We do nothing to use the default bahaviour
		///</summary>
		public void OnRButtonDown(Point pt, ref bool Handled) { }

		///<summary>
		/// OnRButtonUp: Called when the right mouse button is released.  We do nothing to use the default bahaviour, which
		/// is to show the default context menu
		///</summary>
		public void OnRButtonUp(Point pt, ref bool Handled) { }

		///<summary>
		/// OnTimer: Called when a timer event is fired. USe ToolStateAttributes to start and stop timers
		///</summary>
		public void OnTimer(int timerId, int timeEllapsed) { }

		///<summary>
		/// Reset: Helper function to hide the ui and reset the isDrawing flag.
		///</summary>
		private void Reset()
		{
			pts.RemoveAll();
			ui.Hide();
			isDrawing = false;
		}

		private Curve CreateZigZagCurve()
		{
			Curve result = null;
			if (pts.Count > 1)
			{
				result = Application.CreateCurve();

				PointRange smoothedPoints = pts.GetCopy();
				smoothedPoints.Smoothen(3, false);
				Curve stroke = Application.CreateCurve();
				stroke.AppendSubpathFitToPoints(smoothedPoints);
				if (stroke.SubPaths.Count > 0)
				{
					double height = Application.GetApplicationPreferenceValue("ZigZagTool", "Height");
					double distance = Application.GetApplicationPreferenceValue("ZigZagTool", "Distance");

					int divisions = (int)Math.Round(stroke.Length / distance);
					if (divisions > 0)
						stroke.SubPaths.First.EqualDivide(divisions);

					Point first = stroke.SubPaths.First.GetPointAt(0.0);
					SubPath path = result.CreateSubPath(first.x, first.y);

					for (int i = 2; i <= stroke.SubPaths.Count; ++i)
					{
						Vector perpendicular = stroke.SubPaths[i].GetPerpendicularVectorAt(0.0);
						Point position = stroke.SubPaths[i].GetPointAt(0.0);
						perpendicular.Length = ((i & 1) == 1) ? height : -height;
						position.Add(perpendicular);
						path.AppendLineSegment(position.x, position.y);
					}

					Point last = stroke.SubPaths.Last.GetPointAt(1.0);
					path.AppendLineSegment(last.x, last.y);
				}
			}
			return result;
		}

		PointRange pts;												// mouse move points
		OnScreenCurve ui;											// on screen xor'd UI to represent the line we're drawing
		bool isDrawing = false;								// determines if the tool is currently drawing
		ToolStateAttributes currentAttribs;		// save a copy of the state attributes
		Application Application;							// save a copy of the application
	}
}
